#############################################################################
##
#F SortType( <type> )
##
#   Given the component <type> of a jnf-type, this component will be sorted,
#   such that the lists are unique for one type
#
SortType := function( type )
    local b, f, i, G, g, H, l, L;
    b := List( type, x -> x[1] );
    f := List( type, x -> x[2] );
    l := Length(type);
    SortParallel( f, b );
    for i in [1..l] do
        Sort(b[i]);
    od;
    G := DuplicateFreeList(f);
    for g in G do
        H := Filtered( [1..l], x -> f[x]=g );
        L := b{H}; Sort(L); b{H} := L;
    od;
    return List([1..l], x -> [b[x],f[x]]);
end;


#############################################################################
##
#F KronByJNFandPrime( <jnf>, <res>, <n>, <p> )
##
#   Given a jnf-type <jnf> of a matrix of size <n> and its result from
#   'DetermineEqns' <res>, the types of matrices of size n^2 are computed
#   over a field of characteristic <p>.
#   Note: A sufficient large prime can be given as 'p=0'
#
KronByJNFandPrime := function( jnf, res, n, p )
    local K, W, j, V;

    # get type
    K := KroneckerOneType( p, jnf );
    CorrectType( K );

    # adjust subtypes
    W := List( res, x -> false );
    for j in [1..Length(res)] do
        V := CorrectSubtype( K, res[j] );
        V := List( [1..Length(V.degs)], x -> [V.blks[x], V.degs[x]] );
        W[j] := SortType(V);
    od;
   
    return W;
end;


#############################################################################
##
#F Kronecker( <jnfs>, <res>, <n> )
##
Kronecker := function( jnfs, res, n )
    local char, type, kron, i, j, c;
    
    #setup
    char := ELIST[n];
    type := List( [1..Length(jnfs)], 
                  x -> List([1..Length(res[x])], y -> List( char, z -> []  )));
    
    #get the kronecker products
    for i in [1..Length(jnfs)] do
        for c in [1..Length(char)] do
            kron := KronByJNFandPrime( jnfs[i], res[i], n, char[c] );
            for j in [1..Length(res[i])] do          
                type[i][j][c] := kron[j];
            od;
        od;
    od;  
    
    return type;
end;


#############################################################################
##
#F WedgeType( <type> )
##
WedgeType := function( type )
    local mat;
    Error("not yet implemented");
#    mat := CompoundMatType( type, 2 );
#    mat.mult := List( mat.blks, x -> Sum(PartitionChange(x)));
    return mat;
end;


#############################################################################
##
#F AddEigenvalues( q, type )
##
## Let degs = [f1, ..., fd]. Then type has m = f1 + .. + fd different 
## eigenvalues. These lead to m^2 eigenvalues for the kronecker product.
## These function determines the different eigenvalues among the m^2 and
## their multiplicities and add this to the type.
##
AddEigenvaluesKronecker := function( q, type )
    local f, d, I, eigv, i, j, pl, ml, a, k;

    # set up
    f := List(type.type, x -> x[2]);
    d := Length(f);
    I := IdentityMat(d)*q^0;

    # eigenvalues for type
    eigv := [];
    for i in [1..d] do
        for j in [1..f[i]] do
            Add( eigv, I[i]*q^(j-1) );
        od;
    od;

    # eigenvalues for kronecker with multiplicities
    pl := [];
    ml := [];
    for i in [1..Length(eigv)] do
        for j in [1..Length(eigv)] do
            a := eigv[i]+eigv[j];
            k := Position(pl, a);
            if IsBool(k) then
                Add(pl, a);
                Add(ml, 1);
            else
                ml[k] := ml[k]+1;
            fi;
        od;
    od;

    # add info to type
#    type.mult := ml;
    type.eigv := pl;

end;


AddEigenvaluesWedge := function( q, type )
    local f, d, I, eigv, i, j, pl, ml, a, k;

    # set up
    f := List(type.type, x -> x[2]);
    d := Length(f);
    I := IdentityMat(d)*q^0;

    # eigenvalues for type
    eigv := [];
    for i in [1..d] do
        for j in [1..f[i]] do
            Add( eigv, I[i]*q^(j-1) );
        od;
    od;

    # eigenvalues for wedge with multiplicities
    pl := [];
    ml := [];
    for i in [1..Length(eigv)] do
        for j in [1..i-1] do
            a := eigv[i]+eigv[j];
            k := Position(pl, a);
            if IsBool(k) then
                Add(pl, a);
                Add(ml, 1);
            else
                ml[k] := ml[k]+1;
            fi;
        od;
    od;

    # add info to type
#    type.mult := ml;
    type.eigv := pl;

end;


AddEigenvalues := function( q, type, act )
    if NameFunction(act) = "Kronecker" then 
        AddEigenvaluesKronecker(q, type);
    elif NameFunction(act) = "WedgeType" then 
        AddEigenvaluesWedge(q, type);
    else
        Error("not yet implemented");
    fi;
end;

