#############################################################################
##
#F SplitMatrixIntoBlocksExtensionField( <mat> )
##
#   Given a matrix <mat> as the usual record, the function splits <mat> into
#   several block matrices, each block being a single Jordan block.
#   The computation is done using a suitable field extension, such that the
#   degrees (.degs) of the used eigenvalues will always be "1". 
#   When going back to the ground field later on, the component .degs will
#   be changed.
#   The degree of the used (maximal) field extension is stored via .fld.
#
SplitMatrixIntoBlocksExtensionField := function( mat )
    local q, blks, degs, i, j, k, l, n, v;
    
    q    := NameToIndet("q");
    degs := List( mat.type, x -> x[2]);
    l    := Length(degs);
    k    := Sum(List( [1..l], x -> degs[x]*Length(mat.type[x][1])));
    blks := List( [1..k], x -> rec());
    
    k := 0;
    for i in [1..l] do
        for j in [0..degs[i]-1] do
            v    := List( [1..l], x -> 0*q);
            v[i] := q^j;
            for n in mat.type[i][1] do
                k    := k+1;
                blks[k].eigv := [v];
                blks[k].type := [[[n],mat.type[i][2]]];        
            od;    
        od;
    od;
    return blks;
end;

SplitMatrixIntoBlocksExtensionFieldOLD := function( mat )
    local q, blks, degs, i, j, k, l, n, v;
    
    q    := NameToIndet("q");
    degs := List( mat.type, x -> x[2]);
    l    := Length(degs);
    k    := Sum(List( [1..l], x -> degs[x]*Length(mat.type[x][1])));
    blks := List( [1..k], x -> rec());
    
    k := 0;
    for i in [1..l] do
        for j in [0..degs[i]-1] do
            v    := List( [1..l], x -> 0*q);
            v[i] := q^j;
            for n in mat.type[i][1] do
                k    := k+1;
                blks[k].eigv := [v];
                blks[k].type := [[[n],mat.type[i][2]]];        
            od;    
        od;
    od;
    return blks;
end;


#############################################################################
##
#F AppendBlockToMatrix( <mat1>, <mat2> )
##
#   Given the record describing a matrix <mat1> and the record describing 
#   a single(!) Jordan block <mat2>, the record describing the direct sum
#   of both given matrices is returned.
#
AppendBlockToMatrix := function( mat1, mat2 ) 
    local mat, k, b;   
    mat := mat1;
    if not mat2.eigv[1] in mat1.eigv then
        # Case: A block with a new eigenvalue is appended
        Add( mat.eigv, mat2.eigv[1] );
        Add( mat.type, mat2.type[1] );
    else
        # Case: Eigenvalue of new block already exists
        k := Position( mat1.eigv, mat2.eigv[1] );
        b := Concatenation( mat1.type[k][1], mat2.type[1][1]);
        Sort(b);
        mat.type[k] := [b, mat1.type[k][2]];
    fi;       
    return mat;
end; 
