#   The functions within this file take a type <type> of matrices of GL(n,q) 
#   and a set of equations describing a subtype arising of <type> when 
#   applying the Kronecker product.
#   Using this information the type of the subtype in the sense of this 
#   package is computed.

#############################################################################
##
#F MaximalFieldOfEV( <type> )
##
#   For each eigenvalue in <type>.eigv it will be determined from which field
#   extension the factors come. The maximal degree of an extension field such
#   that the eigenvalues of the Kronecker product come from that field is then
#   computed as the LCM of the degrees of the factors used to get the eigen-
#   value.
# 
MaximalFieldOfEV := function( type )
    local i, l, f, q;
    q := NameToIndet("q");
    l := Length(type.eigv);
    type.degs := List( [1..l], x -> 0 );
    
    for i in [1..l] do
        f := Filtered( [1..Length(type.eigv[i])], 
                        x -> not type.eigv[i][x] = 0*type.eigv[i][x]);    
        type.degs[i] := Lcm( type.eigdeg{f} );
    od;
    return;
end;


#############################################################################
##
#F AreGaloisConjugates( <ev1>, <ev2>, <fld> )
##
#   Given two eigenvalues <ev1> and <ev2> by the exponents of a monomial
#   expression, and a list <fld> stating the degree of the field extension
#   where the factors of the eigenvalues are from, it will be returned
#   <true> if <ev1> and <ev2> are Galois conjugated. Otherwise, it will be
#   returned <false>.
#
AreGaloisConjugates := function( ev1, ev2, fld )
    local q, pr, pot, l;

    q := NameToIndet("q");
    
    # Trivial case: if both eigenvalues are identical, then they are
    # Galois conjugated, too.
    if ev1 = ev2 then return true; fi;
    
    # Test, whether occurring zeroes are the same, if not the eigenvalues
    # cannot be Galois conjugates.
    pr := List( [1..Length(ev1)], x -> ev1[x]*ev2[x] );    
    if (not Length(Filtered( ev1, x -> not x=0*q)) = 
            Length(Filtered( pr, x -> not x=0*q))) or
       (not Length(Filtered( ev2, x -> not x=0*q)) = 
            Length(Filtered( pr, x -> not x=0*q))) then
        return false;
    fi;
    
    # Test for all possible field extension, if the eigenvalues are conjugated
    for l in [1..Lcm(fld)] do
        pot := List( [1..Length(fld)], 
                        x -> q^(l mod fld[x] ) mod (q^fld[x]-1));
        pr := List( [1..Length(ev1)], x -> ev1[x]*pot[x] mod (q^fld[x]-1) );
        if pr = ev2 then return true; fi;
    od;
    return false;
end;


#############################################################################
##
#F DetermineConjugates( <type> )
##
#   Given a type <type> representing a type within the GL(r,q), all pairs of 
#   eigenvalues are tested whether or not they are Galois conjugated. 
#   If so, the positions within the list <type>.eigv of conjugated eigenvalues
#   will be stored as a list. These lists will be added to type via
#   the list <type>.conj, which will be appended to <type>.
#
DetermineConjugates := function( type )
    local q, g, l, d, i, j, test;
    q := NameToIndet("q");
    l := Length(type.eigv);
    d := type.eigdeg;
    type.conj := List( [1..l], x -> false );
    
    for i in [1..l] do
        if IsBool(type.conj[i]) and not type.conj[i] then
            for j in [i+1..l] do
                if IsBool(type.conj[j]) then
                    test := AreGaloisConjugates( type.eigv[i], type.eigv[j], d );
                    if test then type.conj[j] := i; fi;
                fi;
            od;
            type.conj[i] := true;
        fi;
    od;  
    if ForAny( type.conj, x -> x = false ) then 
        Error("'DetermineConjugates' produced an error.\n");
    fi;
    return;
end;


#############################################################################
##
#F TestConjugatedEVForField( <type> )
##
#   Note: The record type must already have the component <type>.conj.
#   Note: The field degrees for each eigenvalue must already have been
#         corrected by "MaximalFieldOfEV".
#   
#   For all conjugatd eigenvalues of <type> this functions determines the
#   correct degree of a field extension, from which the eigenvalues come.
#
TestConjugatedEVForField := function( type )
    local con, c, div, d, EV, l, q;
    q := NameToIndet("q");
    
    con := Filtered([1..Length(type.conj)], x -> IsBool(type.conj[x]));
    con := List( con, x -> Concatenation( [x], Positions( type.conj, x )));

    for c in con do 
        if not Length(c) = type.degs[c[1]] then
            div := DivisorsInt( type.degs[c[1]] );
            EV := type.eigv{c}*q^0;
            for d in div do
                l := List( EV, x -> List( x, 
                               y -> y*q^d mod (q^type.degs[c[1]] - 1)) = x );                              
                if ForAll(l, x -> x ) then
                    type.degs{c} := List( c, x -> d);
                    break;
                fi;
            od;
        fi;
    od;
    return;
end;


TestSingleEVForField := function( type )
    local elm, EV, pow, pol, i, q;
    q := NameToIndet("q");
    
    # Get all eigenvalues that are NOT Galois conjugates of any other EV
    elm := Filtered( type.conj, x -> IsInt(x));
    elm := Filtered( [1..Length(type.conj)], 
                      x -> IsBool(type.conj[x]) and not x in elm );                        
    EV  := type.eigv{elm};
    
    # eigdeg is list of degrees belonging to the factors of the EV
    pow := List( type.eigdeg, x -> Sum( List([1..x], y -> q^(y-1))));
    pol := List( EV, x -> List( [1..Length(pow)], y -> x[y]/pow[y] ));
    for i in [1..Length(EV)] do
        if  ForAll( pol[i], y -> IsPolynomial(y) ) then
            type.degs[elm[i]] := 1;
        fi;
    od;
    return;
end;


#############################################################################
##
#F CorrectType( <type> )
##
#   Given a type <type> as result of the function "TypesAfterKron", all 
#   functions mentioned above will be applied to the <type> in a correct
#   order.
#   Some components are added:
#   .conj   A list of the same length as .eigv
#           (Only eigenvalues .eigv[i] with .conj[i]=true must be used later on,
#            .conj[i]=k means that .eigv[i] is Galois conjugates to the k'th EV)
#   .degs   A list of the same length as .eigv
#           (.degs[i] is the maximal a degree of a suitabele field extension
#            such that .eigv[i] comes from that extesnion)
#   .eigdeg A list denoting the eigenvalues 
# 
CorrectType := function( type )
    MaximalFieldOfEV( type );
    DetermineConjugates( type );
    TestConjugatedEVForField( type ); 
    TestSingleEVForField( type );
    return;
end;


#############################################################################
##
#F FieldExtensionOfConjugatedAndEqualEigenvalues( <type>, <subtype>, <pa> )
##
#   Given a subtype <subtype> of a certain type <type> and a list <pa> of
#   positions of eigenvalues, then: 
#   For every pair of eigenvalues in <pa> it will be determined from 
#   which field the eigenvalues come.
#   So, each pair [a,b] will be replaced by [[a,b],d] such that the first 
#   element is the given pair and the second element is the degree of the
#   field extension.
#   All this information will be stored in a list that will be returned.
#
FieldExtensionOfConjugatedAndEqualEigenvalues := function( type, res, pa)
    local q, i, EV, div, d, j, bool, eq, EVmod;
    q := NameToIndet("q");
    eq := [];
    for i in [1..Length(pa)] do
        EV    := List( type.eigv{pa[i]} );
        EVmod := List( EV, x -> x mod (q^type.degs[pa[i][1]] -1) );
        div   := DivisorsInt(type.degs[pa[i][1]]);
        div   := div{[1..Length(div)-1]};
        for d in div do
            bool := List( EV, 
                    x -> x*q^d mod (q^type.degs[pa[i][1]] -1) in EVmod);
            if ForAll( bool, x -> x) then
                Add(eq, [pa[i],d] );
                break;
            fi;
        od;
    od;
    return eq;
end;


CorrectEquations := function( eqns )
    local eq, i, j;
    if ForAll( List( eqns, Length ), x -> x = 1) then return eqns; fi;
    eq := [];
    for i in [1..Length(eqns)] do
        for j in [1..Length(eqns[i])] do
            Add( eq, [eqns[i][j]]);
        od;
    od;
    return eq;
end;


#############################################################################
##
#F CorrectSubtype( <type>, <subtype> )
##
#   Given a subtype <subtype> of a certain type <type>, then the record 
#   <subtype> will be corrected:
#   The eigenvalues will be collected into blocks, such that the
#   corresponding polynomial, which has the collected eigenvalues as roots,
#   is a polynomial over the ground field F_p.
#   The degrees will be determined and the corresponding blocks will be
#   computed. So, <subtype> will be extended by the components 
#   <subtype>.blks and <subtype>.degs
#
CorrectSubtype := function( type, res )
    local q, eqns, eq, eqp, pa, i, j, cjp, con, degs, blks, total;
    q    := NameToIndet("q");
    eqns := res.eqns*q^0;
    eqns := CorrectEquations(eqns);
    eqp  := [];
    
    #Translate EV into positions within list 'type.eigv'
    for eq in eqns do
        pa := [Position( type.eigv, eq[1][1]), Position( type.eigv, eq[1][2])];
        Sort(pa);
        Add(eqp, pa);
    od;
    
    #Collect pairs together
    for i in [1..Length(eqp)] do
        if not IsBool(eqp[i]) then
            for j in [i+1..Length(eqp)] do
                if not IsBool(eqp[j]) and 
                   not Intersection( eqp[i], eqp[j] ) = [] then
                    eqp[i] := Union( eqp[i], eqp[j] );
                    eqp[j] := false;
                fi;
            od;
        fi;
    od;

    if fail in Flat(eqp) then 
        Error("Something went terribly wrong in 'CorrectSubtype'.");
    fi;
   
    #Positions of EV that are identical in subtype
    eqp := Filtered( eqp, x -> not IsBool(x) );
    #Positions of identical EV that are conjugated
    cjp := FieldExtensionOfConjugatedAndEqualEigenvalues ( type, res, eqp );
    #Positions of conjugated EV from type
    con := Filtered([1..Length(type.conj)], x -> IsBool(type.conj[x]));
    con := List( con, x -> Concatenation( [x], Positions( type.conj, x )));

    #Collect all possible EV that are conjugates or identical
    total := List( con, x -> x );
    for i in [1..Length(con)] do
        for j in [1..Length(eqp)] do
            if not Intersection( con[i], eqp[j] ) = [] then
                total[i] := Union( total[i], eqp[j] );
            fi;
        od;
    od;
    total := DuplicateFreeList( total);
   
    #Append to this list all left over EV
    con := [1..Length(type.eigv)];
    SubtractSet( con, Flat(total));
    for i in con do
        Add( total, [i] );
    od;
    
    #Create the list with the degrees
    degs := List( type.degs, x -> 0 );
    for i in [1..Length(total)] do
        for j in [1..Length(total[i])] do
            degs[total[i][j]] := Gcd( type.degs{ total[i] } );
        od;
    od;
    
    #Create the list with the blks
    blks := List( type.type, x -> x[1] );
    for i in [1..Length(eqp)] do
        for j in [2..Length(eqp[i])] do
            blks[eqp[i][1]] := Concatenation( blks[eqp[i][1]],blks[eqp[i][j]]);
        od;
    od;
 
    #Take care of conjugated eigenvalues that are identical
    for i in [1..Length(cjp)] do
        for j in [1..Length(total)] do
            if total[j] = cjp[i][1] then
                degs[cjp[i][1][1]] := cjp[i][2];            
            elif IsSubset(total[j], cjp[i][1]) then
                pa := type.degs{total[j]};
                Add(pa, cjp[i][2]);
                degs[cjp[i][1][1]] := Gcd(pa);                
            fi;
        od;
    od;
    
    #Prepare Output
    total := List( total, x -> x[1] );
    degs := degs{total};
    blks := blks{total};
        
    #Make output
    return rec( degs := degs, 
                blks := blks,
                enum := res.enum, 
                eqns := res.eqns );
end;



CheckTypes := function( types, n )
    local i, j, c, L;
    for i in [1..Length(types)] do
        for c in [1..Length(types[i].acttyp)] do
            for j in [1..Length(types[i].acttyp[c])] do
                L := types[i].acttyp[c][j][1];
                L := List( L, x -> x[2]*Sum(x[1]));
                if not Sum(L) = n^2 then
                    Print("Fehler bei i=",i,", c=",c,", j=",j,"\n",types[i].acttyp[c][j],"\n\n");
                fi;
            od;
        od;
    od;
end;
        
