#############################################################################
##
## KFunctions precomputes the functions k (see Vaughan-Lee, Choosing 
## elements from finite fields, page 6)
##
KFunctions := function(S, m)
    local k, i, iter, T, dT, j;

    # check if this is known
    if IsBound(KLIST[m]) then return KLIST[m]; fi;

    # if not, then compute
    k := List([1..m], x -> 0);
    for i in [1..m] do
        iter := IteratorOfCombinations( S );
        while not IsDoneIterator( iter ) do
            T  := NextIterator( iter );
            dT := Product( T );
            j  := i mod (m/dT);
            if m=dT then 
                k[i] := k[i] + (-1)^Size(T);
            else
                k[i] := k[i] + (-1)^Size(T) * NameToIndet([m/dT,j]);
            fi;
       od;
    od;

    # store and return
    KLIST[m] := k;
    return k;
end;

#############################################################################
##
## GcdOnPoints( list )
##
## This function takes as input a list f_1, ..., f_r of polynomials in a 
## variable q. This function determines a PORC polynomial f in q such that 
## f(a) = Gcd( f_1(a), ..., f_r(a)) for all a in Z. 
##
GcdOnPoints := function( list )
    local pols, coef, gcd, prim, m, d, k, S, f;
    
    # catch a special case
    if Length(list) = 1 then return list[1]; fi;

    # then get the gcd of the polynomials
    coef := GcdRepresentation( list );
    gcd  := Sum( List([1..Length(list)], i -> coef[i]*list[i]));
    prim := PrimitivePolynomial( gcd );
    coef := coef/prim[2];
    gcd  := prim[1];

    # determine Lcm of the denominators of the coefficients
    m := Flat( List( coef, x -> CoefficientsOfUnivariatePolynomial(x) ));
    m := Lcm( List(m, DenominatorRat) );
    
    # catch a special case
    if m=1 then return gcd; fi;

    # get porc-part
    pols := List(list, x -> x/gcd);
    d := List( [1..m], i -> Gcd( List(pols, x -> Value(x,i))));
    if Length(Set(d)) = 1 then return d[1]*gcd; fi;

    # k-functions
    S := PrimeDivisors(m);
    k := KFunctions(S, m);

    # determine linear combination and divisor
    f := m * Product(List(S, x -> (1-1/x)));
    k := Sum(List([1..m], x -> d[x]*k[x]))/f;

    # that's it
    return gcd * k;
end;

GcdOnPoints_Alt := function( list )
    local pols, coef, gcd, prim, m, d, i, k, v, S;
    
    # first try to reduce list
    pols := Filtered(list, x -> x <> 0*x);
    pols := List(pols, x -> x*SignInt(LeadingCoefficient(x)) );
    pols := DuplicateFreeList( pols );

    # then get the gcd of the polynomials
    coef := GcdRepresentation( pols );
    gcd  := Sum( List([1..Length(pols)], i -> coef[i]*pols[i]));
    prim := PrimitivePolynomial( gcd );
    coef := coef/prim[2];

    # determine Lcm of the denominators of the coefficients
    m := Flat( List( coef, x -> CoefficientsOfUnivariatePolynomial(x) ));
    m := Lcm( List(m, DenominatorRat) );
    
    # catch a special case
    if m=1 then return prim[1]; fi;

    # coefficients
    d := List( [1..m], x -> 0 );
    for i in [1..m] do
        k := 0;
        v := Value(gcd,i+k*m);
        while v = 0 do
            k := k+1;
            v := Value(gcd,i+k*m);
        od;
        d[i] := Gcd( List( pols, x -> Value(x, i+k*m)));
        d[i] := d[i]/v;
    od;

    S := PrimeDivisors(m);
    k := KFunctions(S, m);
    k := Sum(List([1..m], x -> d[x]*k[x]));
    return gcd * k/( m * Product(List(S, x -> (1-1/x))));
end;

