#############################################################################
##
#F BlocksOfKroneckerSingleBlocks( <n1>, <n2> )
##
#   Returns the block structure of the Kronecker product of two Jordan blocks
#   of length <n1> and <n2> respectively.
#
BlocksOfKroneckerSingleBlocks := function ( n1, n2 )
    return List( [1..Minimum(n1,n2)], x -> n1+n2-(2*x-1));
end;




#############################################################################
##
#F KroneckerTwoSingleBlocks( <mat1>, <mat2> )
##
#   Given two records <mat1> and <mat2> describing the types of two Jordan 
#   blocks, a record describing the Kronecker product of these two blocks
#   is returned.
#
KroneckerTwoSingleBlocks := function( mat1, mat2 )
    local blks, mat;
    blks := BlocksOfKroneckerSingleBlocks( mat1.type[1][1][1], 
                                           mat2.type[1][1][1] );
    mat  := rec();
    mat.type := [ [blks, 1] ];              
    mat.eigv := [ mat1.eigv[1] + mat2.eigv[1] ];
    return mat;  
end;
KroneckerTwoSingleBlocksOLD := function( mat1, mat2 )
    local blks, degs, mat;
    blks := BlocksOfKroneckerSingleBlocks( mat1.type[1][1][1], 
                                           mat2.type[1][1][1] );
    degs := Lcm( mat1.type[1][2], mat2.type[1][2] );
    mat  := rec();
    mat.type := [ [blks, degs] ];             
    mat.eigv := [ mat1.eigv[1] + mat2.eigv[1] ];
    return mat;  
end;


#############################################################################
##
#F DeterminantalDivisorK( <mat>, <k> )
##
#   Determines the Gcd of all minors of size <k> of the matrix <mat>
#
DeterminantalDivisorK := function( mat, k )
    local l, it1, it2, oO, oP, div;
    l := Length(mat);
    if k = 0 then return mat[1][1]^0; fi;
    if k = l then return Determinant(mat); fi;
    if k > l or k < 0 then Error("Wrong value of 'k'.\n"); fi;
    it1 := IteratorOfCombinations( [1..l], k );
    it2 := IteratorOfCombinations( [1..l], k );
    div := 0*mat[1][1];
    while not IsDoneIterator( it1 ) do
        oO := NextIterator( it1 );
        while not IsDoneIterator( it2 ) do
            oP := NextIterator( it2 );
            div := Gcd( div, Determinant( mat{oO}{oP} ));
        od;
    od;
    return div;
end;



#############################################################################
##
#F ElementaryDivisors( <p>, <q>, <z>)
##
#   Determines the elementary divisors of the qxq-matrix defined as follows:
#                        ( l*Id_q + U_q )^p
#   where U_q denotes the qxq-Matrix with 1's along the first upper diagonal.
#   The matrix is constructed over the field with <z> elements.
#
ElementaryDivisors := function( p, q, z )
    local m, i, l, k, el;
    
    #Setup Matrix
    l := Indeterminate( GF(z), "l" );
    m := l*IdentityMat(q,GF(z));
    for i in [1..q-1] do
        m[i][i+1] := One(GF(z))*l^0;
    od;
    m := Product(List( [1..p], x -> m ));
    
    el := List( [0..q], x -> DeterminantalDivisorK( m, x ));
    el := List( [2..q+1], x -> Degree(el[x]/el[x-1]));
    el := Filtered( el, x -> not x = 0);
    Sort(el);
          
    return Reversed(el);
end;



#############################################################################
##
#F KroneckerTwoSingleBlocksChar( <char>, <mat1>, <mat2> )
##
#   Given two records <mat1> and <mat2> describing a single Jordan block each, 
#   a record describing the Kronecker product of both blocks is returned.
#   This algorithms uses a ground field of characteristic <char>.
#
#   (It is an adoption of the proof of Theorem 1.3 (using the results of 
#   Theorem 1.2) by the book by Marvi Marcus:
#   'Finite Dimensional Multilinear Algebra, Part II, Chapter 7') 
#
KroneckerTwoSingleBlocksChar := function( char, mat1, mat2 )
    local p, q, el, mat;

    p := Maximum( Sum( mat1.type[1][1] ),  Sum( mat2.type[1][1] ));
    q := Minimum( Sum( mat1.type[1][1] ),  Sum( mat2.type[1][1] ));
    
    el := ElementaryDivisors( p, q, char );

    # Create the new record
    mat := rec();
    mat.type := [[el,1]];
    mat.eigv := [mat1.eigv[1] + mat2.eigv[1]];

    return mat;  
end;




#############################################################################
##
#F KroneckerOneType( <char>, <mat> )
##
#   Given a record <mat> describing the type of a matrix, the result of this
#   function is a record describing the Kronecker product of this matrix
#   with itself over a field of characteristic <char>.
#   Within this new record it is assumed that all appearing eigenvalues are
#   distinct. (All equations among those eigenvalues will be considered
#   using subtypes.)
#   Note: For sufficiently large characteristics the Kronecker product behaves
#         like characteristic zero. In this case the computation is much faster.
#
KroneckerOneType := function( char, mat1 )
    local bl, l, i, j, k, mat, eig;
    
    eig := List( mat1.type, x -> x[2] );
    
    #   Split the given matrix into Jordan blocks.
    bl := SplitMatrixIntoBlocksExtensionField( mat1 );     
    l  := Length(bl);

    #   Compute Kronecker product.
    #   First: Compute products of any two single blocks
    mat := List([1..l^2], x -> 0);
    k   := 0;
    if char = 0 then
        for i in [1..l] do
            for j in [1..l] do
                k := k+1;
#Hier wurde etwas geaendert
                mat[k] := KroneckerTwoSingleBlocks( bl[i], bl[j] );
                #mat[k] := KroneckerTwoSingleBlocksChar( 19, bl[i], bl[j] );
            od;
        od;
    else    
        for i in [1..l] do
            for j in [1..l] do
                k := k+1;
                mat[k] := KroneckerTwoSingleBlocksChar( char, bl[i], bl[j] );
            od;
        od;
    fi;
    
    #   Second: Add up all the blocks
    for i in [2..l^2] do
        mat[1] := AppendBlockToMatrix( mat[1], mat[i] );
    od;
    mat[1].eigdeg := eig;
    if IsBound( mat1.ss) then mat[1].ss := mat1.ss; fi;
    return mat[1];
end;   
