
#############################################################################
##
#F PowerSet( l )
##
PowerSet := function(l)
    local comb, cont, lens;
    comb := Combinations([1..l]);
    Sort(comb, function(a,b) return Length(a) < Length(b); end);
    cont := List(comb, x -> List(comb, y -> IsSubset(y,x)));
    lens := List(comb, x -> Length(x));
    return rec( comb := comb, cont := cont, lens := lens);
end;

#############################################################################
##
#F NVec( vec )
##
NVec := function(a)
    local j;
    j := First(a, x -> x <> 0*x);
    if not IsInt(j) then j := LeadingCoefficient(j); fi;
    if j < 0 then return -a; else return a; fi;
end;


#############################################################################
##
#F GaloisGroupOnEigv( q, type )
##
GaloisGroupOnEigv := function( q, type )
    local f, d, a, i, r, l, j, w;

    f := List(type.type, x -> x[2]);
    d := Length(f);
    a := [];
    r := List(f, x -> q^x-1);

     l := [];
     for j in [1..Length(type.eigv)] do
         w := StructuralCopy(type.eigv[j]);
         w := List([1..d], x -> (w[x]*q) mod r[x]);
         l[j] := Position(type.eigv, w);
    od;
    
    return Group(PermList(l), ());
end;

#############################################################################
##
#F SymmGroupOnVecs( q, type )
##
SymmGroupOnVecs := function( q, type )
    local t, d, s, a, A, e, o, S, b, g, l, i, j, w;

    t := type.type;
    d := Length(t);
    s := 1;

    # action on entries
    a := [];
    for e in Set(t) do
        o := Filtered([1..d], x -> t[x] = e);
        S := SymmetricGroup(o);
        Append(a, GeneratorsOfGroup(S));
        s := s * Size(S);
    od;

    # action on vecs
    a := List(a, x -> PermutationMat(x, d));
    A := Group(a, IdentityMat(d));
    A!.ssize := s;
    return A;
end;

ActionOnBlocks := function( H, dm )
    local g, h, l, i, w, j, G;
    g := [];
    for h in GeneratorsOfGroup(H) do
        l := [];
        for i in [1..Length(dm)] do
            w := NVec(dm[i][1]*h);
            j := First([1..Length(dm)], x -> w in dm[x]);
            l[i] := j;
        od;
        Add(g, PermList(l));
    od;
    G := Group(g, ());
    G!.ssize := H!.ssize;
    return G;
end;

#############################################################################
##
#F CreateElementaryEquations( q, type, ineq )
##
CreateElementaryEquations := function( q, type, ineq )
   local e, l, G, c, o, cm, dm, i, t, s, H, hom;

   # set up
   e := type.eigv;
   l := Length(e);

   # get galois group
   G := GaloisGroupOnEigv( q, type );

   # get combis
   c := Combinations([1..l], 2);
   o := Orbits(G, c, OnSets);

   # eliminate and build orbits
   cm := [];
   dm := [];
   for i in [1..Length(o)] do
       t := List(o[i], x -> e{x});
       s := List(o[i], x -> NVec(e[x[1]]-e[x[2]]));
       if not ForAny(s, x -> x in ineq) then 
           Add(cm, t);
           Add(dm, s);
       fi;
   od;

   # consider symmetries
   H := SymmGroupOnVecs( q, type );
   H := ActionOnBlocks( H, dm );

   return rec( symm := H, vecs := dm, eqns := cm);
end;

#############################################################################
##
#F FieldDescription( q, type )
##
FieldDescription := function( q, type )
    local f, d, I, i;
    f := List(type.type, x-> x[2]);
    d := Length(f);
    I := IdentityMat(d)*q^0;
    for i in [1..d] do I[i][i] := q^f[i] - 1; od;
    return I;
end;

#############################################################################
##
#F FieldInequalities( q, type )
##
FieldInequalities := function( q, type )
    local ineq, f, d, I, m, i, e, j, k;

    # set up
    ineq := [];
    f := List(type.type, x -> x[2]);
    d := Length(f);
    I := IdentityMat(d)*q^0;
    m := List( f, MaxDivisorsInt ); 

    # enforce that elements do not fall into subfields
    for i in [1..d] do
        for e in m[i] do
            Add( ineq, (q^e-1)*I[i]);
        od;
    od;

    # enforce that elements with same degree are different
    for i in [1..d] do
        for j in [i+1..d] do
            Add(ineq, I[i] - I[j]);
            if f[i] = f[j] then 
                for k in [1..f[i]-1] do
                    Add(ineq, I[i] - q^k*I[j]);
                od;
            fi;
        od;
    od;

    # that's it
    return ineq;
end;

#############################################################################
##
#F FindPolys( <M>, <t>, <s>, <i>, <sub> )
##
FindPolys := function(M, t, s, i, pows)
    local l, pol, j;

    # get length
    l := pows.lens[i];

    # get polys
    pol := [];
    for j in [1..i-1] do
        if pows.cont[j][i] and pows.lens[j] = l-1 then
            Append(pol, t[j]);
        fi;
    od;
    for j in [1..Length(s)] do
        if s[j] = pows.comb[i] and M[j] <> 0*M[j] then 
            Add(pol, M[j]);
        fi;
    od;

    return Set(pol);
end;

#############################################################################
##
#F ElmNumbers( <equations>, <fielddescription>, <inequations>, <powerset> )
##
ElmNumbers := function( eqns, fdes, ineq, pows )
    local l, e, r, M, s, t, w, i, j, f;

    # set up
    e := Length(ineq);
    l := 2^e;

    # compute mat minors
    r := Concatenation(ineq, eqns, fdes);
    M := MatMinorsPlus(r);

    # catch info
    s := List(M.sets, x -> Intersection(x, [1..e]));
    M := List(M.minors, x -> x*SignInt(LeadingCoefficient(x)));

    # get polys for each case
    t := List([1..l], x -> fail);
    w := List([1..l], x -> fail);
    f := 0;
    for i in [1..l] do
        t[i] := FindPolys(M, t, s, i, pows);
        j := Position(t, t[i]);
        if j < i then 
            w[i] := w[j];
        else
            w[i] := GcdOnPoints(t[i]);
        fi;
        f := f + (-1)^pows.lens[i] * w[i];
    od;

    # simplify and return
    f := SimplifyPorcPoly(f);
    return f;
end;

#############################################################################
##
#F DetermineEqns( <type>, <act> )
##
DetermineEqns := function( type, act )
    local flds, blks, d, n, q, pl, fdes, ineq, pows, cm, dm , H, av, 
          l, rs, os, ps, i, j, tr, k, al, cn, f, a, b, G;

    # set up
    flds := List(type.type, x -> x[2]);
    blks := List(type.type, x -> x[1]);
    d := Length(flds);
    n := Sum(List([1..d], x -> flds[x]*Sum(blks[x]))); 
    q := NameToIndet("q");

    # determine description of fields
    fdes := FieldDescription( q, type );
    ineq := FieldInequalities( q, type );
    pows := PowerSet(Length(ineq));
    Info(InfoCountAlg, 2, "    Got field description");
	
    # determine eigenvalues 
    AddEigenvalues(q, type, act);
    cm := CreateElementaryEquations( q, type, ineq );
    dm := cm.vecs;
    H := cm.symm;
    cm := cm.eqns; 
    Info(InfoCountAlg, 2, "    Got ",Length(dm)," elementary equations");
    Info(InfoCountAlg, 2, "    Got acting group of size ",Size(H));

    # missing: add avoided equations and reduce dm to a set
    av := [];
    #av := AvoidEquations( dm, flds, act );
    #Info(InfoCountAlg, 2, "    Got ",Length(av), " equations to avoid");
	
    # set up
    l := Length(dm);
    rs := List([1..l+1], x -> []);
    os := List([1..l+1], x -> []);

    # init sets
    os[1] := [[[]]];
    rs[1] := [ElmNumbers([], fdes, ineq, pows)];
    ps := [];

    for i in [1..l] do
        Info(InfoCountAlg, 3, "      consider combis of length ",i," of ",l);
        ps := CombinationsAvoid(l, av, ps, i);
        if Length(ps) = 0 then break; fi;

        Info(InfoCountAlg, 3, "      get orbits on ",Length(ps)," points");
        os[i+1] := Orbits(H, ps, OnSets); 

        Info(InfoCountAlg, 3, "      loop over ",Length(os[i+1]), " orbits");
        for j in [1..Length(os[i+1])] do

            # determine element number
            tr := Concatenation( List(os[i+1][j][1], x -> dm[x]) );
            rs[i+1][j] := ElmNumbers(tr, fdes, ineq, pows);

            # check and print info
            if rs[i+1][j] = 0*rs[i+1][j] then 
                Append(av, os[i+1][j]); 
                Info(InfoCountAlg, 4, "        os ",j," has deg ",-infinity);
            else
                tr := DegreeOfPorcPoly(rs[i+1][j]);
                Info(InfoCountAlg, 4, "        os ",j," has deg ",tr);
            fi;
        od;
    od;

    Info(InfoCountAlg, 2, "    Start inclusion-exclusion");
    os := Concatenation(os);
    rs := Concatenation(rs);
    for i in Reversed([1..Length(os)]) do
        for j in [i+1..Length(os)] do
            for k in [1..Length(os[j])] do
                if IsSubset(os[j][k], os[i][1]) then 
                    rs[i] := rs[i] - rs[j];
                fi;
            od;
        od;
    od;
    Info(InfoCountAlg, 2, "    Done");
	
    # eliminate zeros
    al := Filtered([1..Length(rs)], x -> rs[x]<>0*rs[x]);
    rs := rs{al};
    os := os{al};

    # translate result
    f := Product(flds)*H!.ssize;
    for i in [1..Length(rs)] do
        rs[i] := rec( eqns := cm{os[i][1]}, 
                      enum := rs[i]*Length(os[i])/f );
    od;

    # check
    a := Sum( List(rs, x -> x.enum) );
    b := NrElementsOfType(type);
    if a <> b then Error("element number does not fit"); fi;
    
    return rs;
end;


#############################################################################
##
#F SymmetricGroupByType( <jnf>> )
##
#   Given a type <jnf>, is computed which of the eigenvalues can be swapped
#   without changing the type. The result is a symmetric group acting on 
#   the position of the eigenvalues.
#
SymmetricGroupByType := function( jnf )
    local G, r, i, j;
    G := [];
    r := Length( jnf.type );
    for i in [1..r] do
        for j in [i+1..r] do
            if jnf.type[i] = jnf.type[j] then
                Add( G, (i,j) );
            fi;
        od;
    od;
    return Subgroup( SymmetricGroup(r), G );
end;


#############################################################################
##
#F OnEquations( <eqns>, <g> )
##
#   Given a permutation <g> permuting the entries of a list of equation <eqns>
#   the image of <eqns> under <g> is computed.
#
OnEquations := function( eqns, g )
    if eqns = [] then
        return [];
    else
        return List( [1..Length(eqns)], x -> 
                List( [1..Length(eqns[x])], y -> 
                 [Permuted( eqns[x][y][1], g ), Permuted( eqns[x][y][2], g)] ));
    fi;
end;

#############################################################################
##
#F ExtractEqns( <type1>, <type2>, <res> )
##
ExtractEqns := function( type1, type2, res )
    local c, a, b, r, t, i, j, l, G, H1, H2, os;

    c := SymmCount(type1)/SymmCount(type2);
    r := [];
    
    H1 := SymmetricGroupByType(type1);
    H2 := SymmetricGroupByType(type2);
    G := Subgroup( SymmetricGroup(Length(type1.type)), Difference(Set(H1), Set(H2)));
    
    for i in [1..Length(res)] do 
        if res[i].eqns = [] then
            t := StructuralCopy( res[i] );
            t.enum := t.enum*c;
            Add(r, t);
        else
            os := Orbit( G, res[i].eqns, OnEquations );
            os := List(os, z -> 
                       Set( List( z, x -> 
                            Set( List( x, y -> Set(y))) )) );
            os := DuplicateFreeList(os);
            for j in [1..Length(os)] do
                t := StructuralCopy( res[i] );
                t.enum := t.enum*c/Length(os);
                t.eqns := os[j];
                Add(r, t);
            od;
        fi;
    od;
      
    a := Sum(List(r, x -> x.enum));
    b := NrElementsOfType(type2);
    if a <> b then Error("element number does not fit"); fi;
    
    return r;
end;

